<?php

class teamspeak3_query_api {
  protected $sock = null;
  protected $host;
  protected $port;
  protected $user;
  protected $pass;
  protected $logged_in = false;

  /* ---------- class methods ---------- */

  /* decode \? sequences in given scalar string */
  public static function decode($s) {

    return str_replace(
      array("\x5c\x5c", '\/', '\s', '\p', '\a',  '\b',  '\f',   '\n',   '\r',   '\t',  '\v'),
      array("\x5c",     '/',  ' ',  '|',  "\x7", "\x8", "\x12", "\x10", "\x13", "\x9", "\x11"),
      $s);
  }


  /* explode "kw01=value01 kw02=value02|kw10=value10 ..." into array[0..n][kwX]=valueX */
  public static function str2array($line) {

    $list = array();
    foreach (explode('|', trim($line)) as $row) {
      $arr = array();
      foreach (explode(' ', $row) as $record) {
        list($key, $value) = array_pad(explode('=', $record, 2), 2, '');
        $arr[$key] = self::decode($value);
      }
      $list[] = $arr;
    }
    return $list;
  }

  /* ---------- object methods ---------- */

  /* save credentials at construct time */
  public function __construct($host, $port, $user = '', $pass = '') {
    $this->host = $host;
    $this->port = $port;
    $this->user = $user;
    $this->pass = $pass;
  }


  public function __destruct() {
    $this->disconnect();
  }

  /* primitives */

  /* read one line from socket and strip whitespace */
  public function readline() {
    $line = trim(fgets($this->sock));
    $info = stream_get_meta_data($this->sock);
    if ($info['timed_out']) {
      $this->disconnect(false);
      throw new Exception("Connection timed out");
    }
    return $line;
  }


  /* send a command and throw error if unable to send */
  public function sendline($cmd) {

    if (!fputs($this->sock, "$cmd\n")) {
      $err = error_get_last();
      $this->disconnect(false);
      throw new Exception($err['message']);
    }
  }


  /* determine if sock is connected */
  public function is_connected() {
    return is_resource($this->sock);
  }


  /* determine if user is logged in */
  public function is_logged_in() {
    return $this->logged_in;
  }

  /* low level routines */

  /* connect to Teamspeak 3 server */
  public function connect() {

    // 5 seconds connect timeout, and perform own error handling
    set_error_handler(function($errno, $errstr) { return true; }, E_WARNING);
    $this->sock = fsockopen($this->host, $this->port, $errno, $errstr, 5);
    restore_error_handler();
    if (!$this->is_connected())
      throw new Exception("unable to connect to Teamspeak 3 server '{$this->host}' on port {$this->port}: {$errstr}");

    // timeout of 5 seconds for stream reading
    stream_set_timeout($this->sock, 5, 0);

    // check for proper server response. Either "TS3" or "error=..."
    if (($line = $this->readline()) != 'TS3') {
      $this->disconnect(false);
      if (substr($line, 0, 5) == 'error') {
        $result = self::str2array(substr($line, 5))[0];
        throw new Exception("{$result['msg']} {$result['extra_msg']}");
      }
      throw new Exception("Teamspeak 3 server didn't respond properly with 'TS3'");
    }

    // skip welcome message
    $this->readline();
    $this->logged_in = false;
  }


  /* disconnect from Teamspeak 3 server */
  public function disconnect($graceful = true) {

    // try to log out gracefully with 'quit' first
    if ($graceful && $this->is_connected()) {
      try {
        $this->command('quit');
      }
      catch (Exception $e) { } // ignore errors for this call
    }

    if ($this->is_connected())
      fclose($this->sock);

    $this->sock = null;
    $this->logged_in = false;
  }


  /* send a command to the currently connected teamspeak server
     return an array with:
       $result['cmd']=<command>
       $result['raw'][1..n]=<response line 1..n>
       $result['error']['id']=<number>
       $result['error']['msg']=message
       $result['result'][<kw>]=<value>
  */
  public function command($cmd) {

    $result = array('cmd' => $cmd, 'raw' => array(), 'error' => array('id' => -1, 'msg' => 'unknown'));
    $this->sendline($cmd);

    // a command responds with 1 or 2 lines if success:
    // 1 line (optional) with command-specific response, usually keyword=value ... pairs
    // 1 line (always) with result code
    // error id=0 msg=ok
    // or
    // error id=256 msg=command\snot\sfound
    while ($this->is_connected() && !feof($this->sock)) {
      $line = $this->readline();
      $result['raw'][] = $line;

      // save response code
      // 'error' doesn't mean that an error occurred; it contains the response code
      if (substr($line, 0, 5) == 'error') {
        $result['error'] = self::str2array(substr($line, 5))[0];
        break;
      }
    }

    // handle errors that result in immediate connection shutdown

    // not connected any more
    if (!$this->is_connected()) {
      $this->disconnect(false);
      throw new Exception("Connection aborted");
    }

    // no response
    if (count($result['raw']) == 0) {
      $this->disconnect(false);
      throw new Exception("No response: connection closed");
    }

    // flood ban
    if ($result['error']['id'] == 3331) {
      $this->disconnect(false);
      throw new Exception($result['error']['msg']);
    }

    // got valid result: parse and return

    // save result data
    if (count($result['raw']) == 2)
      $result['result'] = self::str2array($result['raw'][0]);

    return $result;
  }

  /* ------------------------------------------------------------------------- */

  /* high level routines */

  /* login to the server */
  public function login() {

    $result = $this->command("login client_login_name={$this->user} client_login_password={$this->pass}");
    if ($result['error']['id'] != 0)
      throw new Exception("unable to login to the Teamspeak server: {$result['error']['msg']}");

    $this->logged_in = true;
  }

  /* select given virtual server */
  public function select_virtualserver($options) {

    $optstr = '';
    foreach ($options as $kw => $val) {
      $optstr .= "$kw=$val ";
    }
    $result = $this->command("use {$optstr}-virtual");
    if ($result['error']['id'] != 0)
      throw new Exception("unable to select virtual server '$opt': {$result['error']['msg']}");
  }


  /* get server list */
  public function serverlist() {

    $result = $this->command("serverlist -short");
    if ($result['error']['id'] != 0)
      throw new Exception("unable to get server list: {$result['error']['msg']}");

    return $result['result'];
  }


  /* get server info */
  public function serverinfo() {

    $result = $this->command("serverinfo");
    if ($result['error']['id'] != 0)
      throw new Exception("unable to get server info: {$result['error']['msg']}");

    // return first item of result, because serverinfo is a simple list, not a list of lists
    return $result['result'][0];
  }

}
